/*
 * Das ist das Programm zur Steuerung der Mystery Box mit dem eingebauten NodeMCU ESP32 Board. Durch das Lösen verschiedener Rätsel kann die Box geöffnet werden. In der aktuellen Version sind vier Rätsel in der richtigen Reihenfolge zu lösen: 
 * 1. Einschalten (Schrödingers Katze = Reed-Schalter): Wenn die Katze auf der richtigen Position steht startet der Microcontroller
 * 2. Codeeingabe über Numpad: Eingabe der Lösung (Entdeckung Amerikas)
 * 3. Drehschalter in vorgegebene Richtungen drehen (1x drücken, 3 Schritte nach rechts, 3 Schritte nach links)
 * 4. Auge (Photowiderstand): Der Photowiderstand muss verdeckt werden, sobald es dunkel genug ist, ist die Aufgabe erfüllt
 */

//Folgende Bibliotheken werden benötigt. Diese können über Werkzeuge --> Bibliotheken verwalten... eingebunden werden. Durch diese Bibliotheken können Teile der Schaltung (Servo, Numpad und Drehschalter) einfacher programmiert werden.
#include <ESP32Servo.h>
#include <Keypad.h>
#include <Encoder.h>

#define ARRAY_LEN(arr) sizeof(arr)/sizeof(arr[0])

//Hier wird die Pinbelegung am Microcontroller festgelegt und dem jeweiligen Pin ein Name zugewiesen.
namespace PinMap {
  enum Mapping {
    PHOTORESISTOR = 32,
    NUMPAD_COL1 = 33,
    NUMPAD_COL2 = 25,
    NUMPAD_COL3 = 26,
    NUMPAD_ROW1 = 27,
    NUMPAD_ROW2 = 14,
    NUMPAD_ROW3 = 12,
    NUMPAD_ROW4 = 13,
    ENCODER_CLK = 35,
    ENCODER_DATA = 34,
    ENCODER_BUTTON = 23,
    SERVO_SIGNAL = 15,
    SERVO_POWER = 4,
    LED_1 = 5,
    LED_2 = 18,
    LED_3 = 19,l
    LED_4 = 22
  };
};

//Das Numpad ist in 3 Spalten und 4 Reihen eingeteilt. (COL = "column" also Spalte, ROW = "row" also Reihe)
const byte N_COL = 3;
const byte N_ROW = 4;

//Die hexaKey beschreiben, an welcher Stelle sich welches Symbol befindet (in umgekehrter Reihenfolge). Man weiß dadurch, an welcher Position also Spalte x, Reihe y, sich welches Symbol befindet.
char hexaKeys[N_ROW][N_COL] = {
  {'#', '0', '*'},
  {'9', '8', '7'},
  {'6', '5', '4'},
  {'3', '2', '1'}
};

//colPin uns rowPins legt die Pinbelegung des Numpads am Mikrocontroller fest. colPins sind die Pins für die Spalten, rowPins für die Reihen.
byte colPins[N_COL] = {
  PinMap::NUMPAD_COL1, PinMap::NUMPAD_COL2, PinMap::NUMPAD_COL3
};
byte rowPins[N_ROW] = {
  PinMap::NUMPAD_ROW1, PinMap::NUMPAD_ROW2, PinMap::NUMPAD_ROW3, PinMap::NUMPAD_ROW4
};

const byte N_LED = 4; //Anzahl der LEDs, bzw. nächster Block: Pinbelegung für LED_1 bis LED_4
const byte LED_PINS[N_LED] = {
  PinMap::LED_1, PinMap::LED_2, PinMap::LED_3, PinMap::LED_4
};

const int SERVO_ROTATION_ANGLE_INIT = 0; //Die Ausgansposition des Servomotors zur Öffnung der Box
const int SERVO_ROTATION_ANGLE_UNLOCK = 90; //Wenn sich der Servomotor an diese Position (diesen Winkel) bewegt, öffnet sich die Box
const int SERVO_MOVE_TIME_MS = 1000;
const int MAX_ADC_VAL = 4095; //Der Maximalwert des Photowiderstandes.
const int PHOTORESISTOR_LOW_LIGHT_THRESHOLD = 1000; //Dieser Wert legt fest, ab welchem Wert die Aufgabe 4 gelöst ist. Der Wert kann so verändert werden, dass bei mehr (höherer Wert) oder weniger Licht (niedrigerer Wert) geschaltet wird.

Keypad numpad(makeKeymap(hexaKeys), rowPins, colPins, N_ROW, N_COL); //Mit dieser Zeile wird ein "virtuelles" Numpad angelegt, welches mit unserer Konfiguration übereinstimmt, also die Anzahl der Reihen und Spalten in Kombination mit der Pinbelegung.
Encoder encoder(PinMap::ENCODER_DATA, PinMap::ENCODER_CLK); //Ähnlich wie für das Numpad wird hier der Drehschalter (Encoder) angelegt.
Servo servo; //Und als letztes noch der Servomotor

/* 
 * Der folgende Code ist in einzelne Funktionen (also Unterprogramme) gegliedert, welche immer eine gewisse Aufgabe ausführen. Der Name der Funktion (z.B. waitCoverEye) beschreibt grob um welche Aufgabe es sich handelt.
 * Diese Funktionen können dann im Hauptprogramm loop() aufgerufen werden und "ihre" Aufgaben erledigen.
 * 
 * Dazu werden drei wichtige Programmierelemente benötigt:
 * - while-Schleife:  Die while-Schleife wiederholt den in den geschwungenen Klammern {} befindlichen Code solange, bis die Bedingung in den runden Klammern () erfüllt ist. Bei jedem Durchlauf wird die Bedinung überprüft und der nur ausgeführt,
 *                    wenn sie noch zutrifft. Beispiel: while(x < 10){ ... } führt den angegebenen Code solange aus, bis x den Wert 10 erreicht hat.
 *                    Erklärung mit Video: https://studyflix.de/informatik/while-schleifen-in-c-687
 * - for-Schleife:    In den runden Klammern der for-Schleife werden eine Startbedinung, eine Endbedingung und die Schrittweite angegeben. Bei jedem Durchlauf wird dann die Zählervariable um die Schrittweite erhöht, solange bis die Endbedingung
 *                    erfüllt ist. Beispiel: for (x = 0; x < 10; x++){ ... } Die Schleife startet mit x = 0, führt den in den geschwungenen Klammern stehenden Code aus und erhöht dann x um 1 (x++ bedeutet x = x + 1). Wenn x den Wert 10 erreicht,
 *                    wird die Schleife abgebrochen. Der Wert von x wird ebenso im darin befindlichen Code verändert. Wenn also x im Code vorkommt, nimmt es bei jedem Durchlauf auch den entsprechenden Wert im Programmcode an.
 *                    Erklärung mit Video: https://studyflix.de/informatik/for-schleifen-in-c-686
 * - if-Verzweigung:  Die If-Verzweigung überprüft, ob die in den runden Klammern angegebene Bedingung wahr ist. Wenn ja, wird der Code in den geschwungenen Klammern aufgeführt, sonst übersprungen.
 * 
 * Für die Erklärung des Codes bzw. einzelnen Funktionen und Elementen werden Kommentare verwendet. Diese erkennt man entweder an den zwei Schrägstrichen // für einzeilige Kommentare oder an /* für mehrzeilige Kommentare. Außerdem werden Kommentare
 * in grauer Schrift angezeigt wodurch sie sich vom restlichen Code unterscheiden. Kommentare werden beim Ausführen des Programmes ignoriert und können deshalb überall platziert werden um etwas zu erklären.                     
 */

/*
 * Die Numpad-Funktion braucht als sogenannten Übergabeparamter den gewünschten Code. Das erkennt man an der Angabe "const char* code" in der Klammer. Dadurch kann man beim Aufruf der Funktion den gewünschten Code festlegen. Dies erfolgt erst
 * später im Programm.
 * In einer while-Schleife wird, sobald eine Taste gedrückt wird, diese wieder am seriellen Monitor ausgegeben. Dann wird überprüft, ob das eingegebene Symbol richtig ist. Wenn nicht, wird von vorne gestart. Wenn das Zeichen richtig war, wird
 * die nächste Eingabe abgewartet und wiederum überprüft. Wenn am Ende die richtige Zeichenfolge eingegeben wurde, ist die Aufgabe erledigt und das Programm springt an die nächste Stelle.
 */
void waitNumpadCode(const char* code)
{
  const int code_len = strlen(code);
  int next_key_index = 0;
  while (next_key_index < code_len) {
    const char key = numpad.getKey();
    const bool key_pressed = key != 0;
    if (key_pressed) {
      Serial.println(key);
      if (key == code[next_key_index]) {
        ++next_key_index;
      } else {
        next_key_index = 0;
      }
    }
    delay(1);
  }
}

/*
 * Das SafeLockingWheel ist der an der Seite befindliche Drehschalter (Encoder). Dieser wird über die folgende Funktion gesteuert. Die übergebenen Parameter "sequence" und "sequence_len" dienen zur Beschreibung der gewünschten Eingabe. 
 * "sequence" ist die einzugebende Drehfolge und "sequence_len" ist die Länge der gesamten Abfolge. Die genauere Erklärung erfolgt später.
 * Mithilfe des Buttons wird kann der Drehschalter zurückgesetzt werden, wenn eine falsche Eingabe erfolgt ist. Dies wird im ersten if-Block überprüft und im seriellen Monitor ausgegeben.
 * Der zweite if-Block überprüft, ob die eingegebene Sequenz mit der gewünschten Sequenz übereinstimmt. Also z.B. 3x nach rechts (+3) und 3x nach links (-3).
 */
void waitSolveSafeLockingWheel(const int* sequence, const int sequence_len)
{
  encoder.write(0);

  int next_index = 0;
  const int initial_rest_time = 1000;
  int rest_time = initial_rest_time;

  while (next_index < sequence_len) {
    const bool button_pressed = !digitalRead(PinMap::ENCODER_BUTTON);
    const int N_COUNT_PER_STEP = 4;
    const long value = encoder.read() / N_COUNT_PER_STEP;
    if (button_pressed) {
      Serial.println("Button press -> reset locking wheel");
      next_index = 0;
      encoder.write(0);
      rest_time = initial_rest_time;
    }
    if (value == sequence[next_index]) {
      if(rest_time <= 0 ) {
        Serial.println( sequence[next_index]);
        encoder.write(0);
        ++next_index;
      }
      rest_time -= 1;
    } else {
       rest_time = initial_rest_time;
    }
    delay(1);
  }

}

/* 
 * Die Funktion waitCoverEye() dient zur Überprüfung der Helligkeit am Photowiderstand (also im Auge). Dazu wird laufend in einer while-Schleife überprüft, ob der eingelesene Wert (ad_value) größer ist als der "Grenzwert" (PHOTORESISTOR_LOW_LIGHT_THRESHOLD, wird zu Beginn des Programmes festgelegt).
 * Solange das der Fall ist, wird der Wert weiterhin über die Funktion analogRead am entsprechenden Pin eingelesen und über den Seriellen Monitor (Symbol rechts oben in der Arduino IDE) ausgebenen. Wenn der Wert unter den Grenzwert fällt, ist
 * die Funktion beendet und die Aufgabe gelöst.
 */
void waitCoverEye()
{
  int ad_value = MAX_ADC_VAL;
  while ( ad_value > PHOTORESISTOR_LOW_LIGHT_THRESHOLD ) {
    ad_value = analogRead(PinMap::PHOTORESISTOR);
    Serial.println(ad_value);
    delay(100);
  }
}

/*
 * servoPowerup und servoPowerdown dienen nur dazu, den Servomoter ein- bzw. auszuschalten. Nur wenn er eingeschaltet ist, kann er auch bewegt werden.
 */
void servoPowerup()
{
  digitalWrite(PinMap::SERVO_POWER, HIGH) ;
}

void servoPowerdown()
{
  digitalWrite(PinMap::SERVO_POWER, LOW) ;
}

/*
 * Hier wird der Servomotor initialisert. Dazu wird er erst eingeschaltet, also mit Strom versorgt, und dann auf die gewünschte Anfangsposition (SERVO_ROTATION_ANGLE_INIT) gestellt. Dadurch wird sichergestellt, dass sich der Motor nach dem Starten der
 * Box an der richtigen Winkelposition befindet. Sobald das erfolgt ist, wird er (vorübergehend) wieder ausgeschaltet.
 */
void servoInitLock()
{
  servoPowerup();
  servo.write(SERVO_ROTATION_ANGLE_INIT);
  delay(SERVO_MOVE_TIME_MS);
  servoPowerdown();
}

/*
 * Diese Funktion wird ausgeführt, wenn alle Rätsel gelöst worden sind. Damit wird also die Box geöffnet. Dazu wird der Servomotor auf den richtigen Winkel gestellt um den Öffnungsmechanismus zu betätigen und dann wieder an den Ausgangspunkt zurückgefahren.
 */
void servoUnlockBox()
{
  servoPowerup();
  servo.write(SERVO_ROTATION_ANGLE_UNLOCK);
  delay(SERVO_MOVE_TIME_MS);
  servoInitLock();
}

/*
 * Funktion zur Initialisierung der LED-Statusleiste. Die LED-Pins am Mikrocontroller werden als Ausgänge festgelegt und die ausgeschaltet. Die for-Schleife dient dazu, die LEDs der Reihe nach zu konfigurieren:
 * Beim Start der Schleife wird led_pins[0] verwendet, das entspricht dem ersten LED. Beim nächsten Durchlauf wird i um eins erhöht, das heißt das dann led_pins[1] verändert wird und so weiter.
 * Die Zuweisung der Pins erfolgt bereits weiter oben im Programm. Die Übergabeparameter sind die festgelegte Pins (led_pins) und die Anzahl der LEDs (n_led).
 */
void ledProgressBarInit(const byte* led_pins, const int n_led)
{
  for (int i = 0; i < n_led; ++i) {
    pinMode(led_pins[i], OUTPUT);
    digitalWrite(led_pins[i], LOW);
  }
}

/*
 * ledProgressBar und ledFun dienen zur Steuerung der LEDs. Der Code ist so gestaltet, dass jederzeit mehr LEDs angeschlossen und gesteuert werden können. Mit "percentage" wird festgelegt, wie viele Aufgaben schon
 * gelöst wurden. Bei vier Aufgaben wären das also pro Aufgabe 25%. Pro 25% wird dann eine weitere LED eingeschaltet. Also die erste bei 25, die zweite bei 50, die dritte bei 75 und die letzte bei 100%.
 * ledFun dient zusätzlich dazu, beim Starten der Box eine kurze LED-Animation anzuzeigen. Dabei werden die LEDs der Reihe nach eingeschaltet. LED4 blinkt dann am Ende noch 3x (n_blink).
 * Die erste for-Schleife schaltet ein LED nach dem anderen ein mit einer Verzögerung von 20ms. Die Verzögerung kann mit delay erzeugt werden. Sobald die Funktion delay aufgerufen wird, bleibt das Programm für die angegebene 
 * Zeit (in Millisekunden) "stehen", also im aktuellen Zustand.
 * Die zweite for-Schleife lässt dann die letzte LED 3x blinken. Die dritte for-Schleife ist ähnlich wie die erste, wird jedoch in umgekehrter Reihenfolge ausgeführt und die LEDs ausgeschaltet.
 */
void ledProgressBar(const byte* led_pins, const int n_led, const int percentage)
{
  const int first_led_off_index = percentage * n_led / 100;
  for (int i = 0; i < n_led; ++i) {
    const int led_on = i < first_led_off_index;
    digitalWrite(led_pins[i], led_on);
  }
}

void ledFun(const byte* led_pins, const int n_led)
{
  for (int i = 100 / n_led; i <= 100; ++i) {
    ledProgressBar(led_pins, n_led, i);
    delay(20);
  }
  const int n_blink = 3;
  for (int i = 0; i < n_blink; ++i) {
    digitalWrite(led_pins[n_led - 1], LOW);
    delay(100);
    digitalWrite(led_pins[n_led - 1], HIGH);
    delay(100);
  }
  for (int i = 100; i > 100/n_led-1; --i) {
    ledProgressBar(led_pins, n_led, i);
    delay(10);
  }
}

/*
 * Die setup-Funktion wird speziell bei Arduino-Boards verwendet. Hier können alle wichtigen Konfiguration festgelegt werden, welche beim Programmstart ausgeführt werden sollen.
 * In unserem Fall werden die Baud-Rate (in diesem Programm keine Erklärung, bei Interesse bitte Google befragen :) ) und der Servo-Pin "angeschlossen". Danach werden noch die Pin-Modi für den Servo (als Output, also Ausgang) 
 * und den Drehschalter-Button (als Input) festgelegt. Zusätzlich werden die beiden Initialisierungsfunktionen für den Servo und die LED-Statusleiste aufgerufen.
 */
void setup() {
  Serial.begin(115200);
  servo.attach(PinMap::SERVO_SIGNAL);
  pinMode( PinMap::SERVO_POWER, OUTPUT);
  pinMode(PinMap::ENCODER_BUTTON, INPUT_PULLUP);

  servoInitLock();
  ledProgressBarInit(LED_PINS, N_LED);
}

/*
 * loop() ist sozusagen das Hauptprogramm. Loop bedeutet soviel wie Schleife und heißt, das alles darin immer wieder ausgeführt wird, solange unser Board mit Strom versorgt wird. Hier wird auch die Reihenfolge für die Rätsel
 * festgelegt. Die Erklärung erfolgt hier direkt bei den einzelnen Codezeilen.
 */
void loop() {
  ledFun(LED_PINS, N_LED); //sobald die Box eingeschaltet wird, wird die LED-Animation ausgeführt
  ledProgressBar(LED_PINS, N_LED, 25); //Da das erste Rätsel (das Einschalten) gelöst ist, leuchtet bereits die erste LED. 25% des Fortschritts sind erreicht.
  waitNumpadCode("1492"); //Hier kann ein beliebiger Code eingegeben werden. Die Funktion waitNumpadCode prüft dann ob der festgelegt Code (bei uns "1492" = Entdeckung Amerikas) richtig eingegeben wurde. Hier kannst du auch einen eigenen Code festlegen :)
  ledProgressBar(LED_PINS, N_LED, 50); //Sobald der Code richtig eingegeben wurde, sind 50% erreicht und LED2 schaltet sich ein.
  const int lock_sequence[] = {3, -3}; //die lock_sequence kann hier festgelegt werden. Laut dem Rätsel auf der Box 3x nach Osten, 3x nach Westen: das entspricht also den Werten +3 und -3. Hier können auch eine eigene Sequenz festgelegt werden. Ein positiver Wert bedeutet im Uhrzeigersinn, ein negativer dagegen.
  waitSolveSafeLockingWheel(lock_sequence, ARRAY_LEN(lock_sequence)); //Die Funktion waitSolveSafeLockingWheel wird mit der vorhin angegebenen Sequenz aufgerufen. Bitte langsam am Rad drehen und bei falscher Eingabe den Button drücken (also einfach auf das Rad drücken).
  ledProgressBar(LED_PINS, N_LED, 75); //Wenn richtig gedreht wurde, sind 75% gelöst und LED3 leuchtet.
  waitCoverEye(); //Für die letzte Aufgabe muss der Photowiderstand (Helligkeitssensor) soweit abgedunkelt werden, dass der Grenzwert (dieser kann ebenfalls verändert werden, wurde weiter oben im Code schon gezeigt) erreicht wird.
  ledProgressBar(LED_PINS, N_LED, 100); //Wenn das erledigt ist, sind alle aufgaben gelöst, LED4 leuchtet auf und
  servoUnlockBox(); //die Box öffnet sich. Gratuliere! :)
}
